/*
** Module   : DatabaseAuthenticationHook.java
** Abstract : Database authentication hook that asks the user for credentials for accessing a
**             database.
**
** Copyright (c) 2005-2010, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ------------------------------- Description -------------------------------
** 001 OM  20130927 First commit with minimum implementation.
*/ 

package com.goldencode.p2j.security;

import com.goldencode.p2j.persist.ConnectionManager;
import com.goldencode.p2j.security.SecurityManager;
import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.ThinClient;
import com.goldencode.p2j.ui.client.OutputManager;
import com.goldencode.p2j.util.*;

import java.io.ByteArrayInputStream;
import java.io.DataInputStream;
import java.io.IOException;
import java.util.*;

import static com.goldencode.p2j.ui.LogicalTerminal.ALERT_MESSAGE;
import static com.goldencode.p2j.ui.LogicalTerminal.BTN_OK;

/**
 * This class is a database authentication hook implementation. If the database requires
 * authentication (ie. the _User meta-table is populated) this hook is used to prompt the user
 * for a userid/password combination. If the credentials cannot be validated, the user is
 * informed and allows to enter the combination again for maximum three times; after that the
 * session is closed.
 * However, the user is allowed to [Cancel] the authorization process and continue as "blank"
 * user (if the database allows this.).
 *
 * ┌─────────────────── Login ───────────────────┐
 * │                                             │
 * │ Please enter a User Id and Password for     │
 * │ database: p2j_test                          │
 * │                                             │
 * │    User Id: ________________                │
 * │   Password: ________________                │
 * │                                             │
 * │              <OK>   <Cancel>                │
 * └─────────────────────────────────────────────┘
 *
 *        ┌──────────── Error ────────────┐
 *        │ Userid/Password is incorrect. │
 *        │ ───────────────────────────── │
 *        │             <OK>              │
 *        └───────────────────────────────┘
 */
public class DatabaseAuthenticationHook
implements  Authenticator // , InitTermListener
{
   /** Edit frame ID. */
   private static final int EDIT_FID = 100;

   /** ID for message on 1st line. */
   private static final int MSG1_ID = EDIT_FID + 2; // 1 + skip

   /** ID for message on 2nd line. */
   private static final int MSG2_ID = EDIT_FID + 4;

   /** User ID field ID. */
   private static final int USER_ID = EDIT_FID + 6;

   /** Password field ID. */
   private static final int PASSWORD_ID = EDIT_FID + 8;
   
   /** OK button field ID. */
   private static final int BUTTON_OK_ID = EDIT_FID + 10;
   
   /** Cancel button field ID. */
   private static final int BUTTON_CANCEL_ID = EDIT_FID + 11;

   /** Keep the id of the last component. */
   private static final int LAST_CONFIG_COMP = EDIT_FID + 12; /* including skip(1) after buttons*/

   
   /** First line of the [Login] dialog. */
   private static final String STD_MSG_1 = "Please enter a User Id and Password for";

   /** First line of the [Login] dialog. The name of teh database will be concatenated. */
   private static final String STD_MSG_2 = "database: ";

   /** The text on OK button from the [Login] dialog */
   private static final String DLG_BTN_OK_TEXT = "OK";

   /** The text on OK button from the [Login] dialog */
   private static final String DLG_BTN_CANCEL_TEXT = "Cancel";

   /** Invalid user status message. */
   private static final String INVALID_CREDENTIALS = "Userid/Password is incorrect.";

   /** The text on OK button from the [Error] dialog */
   private static final String ERR_BTN_OK_TEXT = "OK";

   /** Client interface. */
   private ThinClient client = null;

   /** Edit frame definition. */
   private ScreenDefinition editsd = null;

   /** 
    * The screen buffer for the editing frame.
    * NOTE: numWidgets is important and must be exactly the numbers of widget configs defined.
    */
   private ScreenBuffer editsb = new ScreenBuffer(EDIT_FID, LAST_CONFIG_COMP - EDIT_FID - 1);

   /** The name of the current database that requests authentication. */
   private String database = null;
   
   /**
    * The default constructor.
    */
   public DatabaseAuthenticationHook()
   {
      client = ThinClient.getInstance();
   }

   /**
    * Implements client side authentication logic.
    *
    * @param    parameters
    *           Additional configuration parameters.
    * @param    code
    *           The result of the most recent attempt to authenticate or
    *           <code>AUTH_RESULT_NONE</code> if this is the first attempt.
    *
    * @return   Array of bytes to be transmitted to the server as the authorization input.
    */
   @Override
   public byte[] clientAuthHook(Map<String, Object> parameters, int code)
   {
      OutputManager tk = OutputManager.instance();
      database = (String) parameters.get("database");

      switch (code)
      {
         case AUTH_RESULT_NONE:
            //AUTH_REQ_DB_QUERY

            // push the frame definition the first time in:
            createFrameDefinition(database);
            break;

         case AUTH_RESULT_INVALID_PASSWORD:
            // display error message as box:

            client.messageBox(new Object[]{ INVALID_CREDENTIALS }, null,
                              ALERT_MESSAGE, BTN_OK, " Error ", null);
            break;
      }

      String id = null;
      String pw = null;
      EventList el = new EventList(true);
      ScreenBuffer sbf = null;

      while (true)
      {
         // drive the logon frame I/O
         tk.setInvalidate(true);
         int[] activeWidgets = new int[] { USER_ID, PASSWORD_ID, BUTTON_OK_ID, BUTTON_CANCEL_ID };
         client.enable(EDIT_FID, activeWidgets, editsb, true);
         client.clear(EDIT_FID, true);
         client.view(EDIT_FID, editsb, null, true);
         editsb.resetChanged();
         tk.setInvalidate(false);

         try
         {
            client.setAuthMode(true);
            sbf = client.waitFor(el, USER_ID, -1, editsb);
         }
         catch (ConditionException cx)
         {
            // EndConditionException: END key function executed
            return new byte[] {};
         }
         finally
         {
            client.setAuthMode(false);
         }

         tk.setInvalidate(true);
         client.enable(EDIT_FID, null, editsb, false);
         tk.setInvalidate(false);

         // inspect the input; exit out if requested
         if (Keyboard.keyFunction(sbf.getKeyCode()).equals("END-ERROR"))
         {
            return new byte[] {};
         }

         character idc = (character) sbf.getWidgetValue(USER_ID);
         id = (idc == null) ? "" : idc.getValue().trim();

         character pwc = (character) sbf.getWidgetValue(PASSWORD_ID);
         pw = (pwc == null) ? "" : pwc.getValue();

         break;
      }

      // return a byte array
      return SecurityManager.packageIdPassword(id, pw);
   }

   private void createFrameDefinition(String database)
   {
      // create the frame definition on the fly
      editsd = new ScreenDefinition(EDIT_FID);

      // edit frame configuration
      FrameConfig frameConfig = new FrameConfig(ComponentConfig.FRAME, null);
      frameConfig.setRow(10);
      frameConfig.setBox(true);
      frameConfig.setCentered(true);
      frameConfig.setOverlay(true);
      frameConfig.setDcolor(Color.NORMAL);
      frameConfig.setSideLabels(true);
      frameConfig.setTitle(" Login ");
      frameConfig.setWidthChars(47);
      editsd.addConfig(frameConfig, EDIT_FID);

      // set next widget position
      addSkip(editsd, 1); // id = EDIT_FID + 1

      ControlTextConfig mess1Fic = new FillInConfig(ComponentConfig.FILL_IN, null);
      mess1Fic.setDataType("character");
      mess1Fic.setColumn(2);
      mess1Fic.setStatic(true);
      editsd.addConfig(mess1Fic, MSG1_ID);

      // set next widget position
      addSkip(editsd, 0); // id = EDIT_FID + 3

      ControlTextConfig mess2Fic = new FillInConfig(ComponentConfig.FILL_IN, null);
      mess2Fic.setDataType("character");
      mess2Fic.setStatic(true);
      mess2Fic.setColumn(2);
      editsd.addConfig(mess2Fic, MSG2_ID);

      // set next widget position
      addSkip(editsd, 1);  // id = EDIT_FID + 5

      // user ID entry field
      FillInConfig userFic = new FillInConfig(ComponentConfig.FILL_IN, null);
      userFic.setDataType("character");
      userFic.setFormat("x(16)");
      userFic.setColumn(4);
      userFic.setLabel(" User ID");
      editsd.addConfig(userFic, USER_ID);

      // set next widget position
      addSkip(editsd, 0);  // id = EDIT_FID + 7

      // password entry field
      FillInConfig passFic = new FillInConfig(ComponentConfig.FILL_IN, null);
      passFic.setDataType("character");
      passFic.setFormat("x(16)");
      passFic.setColumn(4);
      passFic.setLabel("Password");
      passFic.setBlank(true);  // this is a password; should not be visible.
      editsd.addConfig(passFic, PASSWORD_ID);

      // set next widget position
      addSkip(editsd, 1);  // id = EDIT_FID + 9

      ButtonConfig bOk = new ButtonConfig();
      bOk.setType(ComponentConfig.BUTTON);
      bOk.setAutoGo(true);
      bOk.setColumn(15);
      bOk.setLabel("OK");
      editsd.addConfig(bOk, BUTTON_OK_ID);

      ButtonConfig bCancel = new ButtonConfig(); 
      bCancel.setType(ComponentConfig.BUTTON);
      bCancel.setAutoEndKey(true);
      bCancel.setColumn(22);
      bCancel.setLabel("Cancel");
      editsd.addConfig(bCancel, BUTTON_CANCEL_ID);

      addSkip(editsd, 1);  // id = EDIT_FID + 12

      // instantiate screen data
      client.pushScreenDefinition(new ScreenDefinition[] { editsd });

      editsb.putWidgetValue(MSG1_ID, new character(STD_MSG_1));
      editsb.putWidgetValue(MSG2_ID, new character(STD_MSG_2 + database));
   }

   /**
    * Implements server side authentication logic.
    * <p>
    * Accepts the byte array produced by the client side authentication hook
    * as authentication input as well as any custom parameters.
    *
    * @param    auth
    *           The authorization input from the client.
    * @param    parameter
    *           Additional configuration parameters taken from the directory.
    *
    * @return   The result of the authentication processing.
    */
   @Override
   public AuthenticationResponse serverAuthHook(byte[] auth, String database)
   {
      if (auth == null)
      {
         // user cancelled the authentication for this database
         // authenticate with 'blank' user: mso.setUserId("", "", database)
         return new AuthenticationResponse(null, AUTH_RESULT_CANCEL_CURRENT);
      }
      // parse the byte array input
      ByteArrayInputStream bis = new ByteArrayInputStream(auth);
      DataInputStream dis = new DataInputStream(bis);

      String userId = null;
      String password = null;

      try
      {
         userId   = dis.readUTF();
         password = dis.readUTF();
      }
      catch (IOException ioe)
      {
         return new AuthenticationResponse(null, AUTH_RESULT_UNSPECIFIED_FAILURE);
      }

      MetadataSecurityOps mso = new MetadataSecurityOps();
      if (mso.setUserId(userId, password, database))
      {
         ConnectionManager.setConnectDbUser(database);
         return new AuthenticationResponse(userId, AUTH_RESULT_SUCCESS);
      }
      else
      {
         return new AuthenticationResponse(null, AUTH_RESULT_INVALID_PASSWORD);
      }
   }

   /**
    * Finalizes any resources allocated during authentication by the client.
    */
   @Override
   public void clientFinalize()
   {
      client.statusInputRevert();
      client.hideAll(true);
      client.destroyFrame(EDIT_FID);
   }

   /**
    * Helper method, adds specified number of empty lines to a ScreenDefinition.
    *
    * @param   sd
    *          The frame to which to add a skip.
    * @param   lines
    *          Lines to skip.
    */
   private static void addSkip(ScreenDefinition sd, int lines)
   {
      SkipConfig sc = new SkipConfig(ComponentConfig.SKIP, null);
      sc.setVertical(true);   // skip not space
      sc.setHeight(lines);
      sd.addConfig(sc);
   }
}
